function vernval_emgr()
% vernval_emgr (chapter 5 emgr: verification and validation)
% by Christian Himpe, 2015-2016 ( wwwmath.uni-muenster.de/u/himpe )
% released under BSD 2-Clause License ( opensource.org/licenses/BSD-2-Clause )
%*
    addpath('../')

    rand('seed',1009);
    randn('seed',1009);

%% SETUP

    J = 10;
    N = 100;
    O = J;
    T = [0.01,1.0];
    L = floor(T(2)/T(1)) + 1;

    X = zeros(N,1);
    U = [ones(J,1)./T(1),zeros(J,L-1)];
    P = N;

    [A,B,C] = ilp(J,N,O,1,1009);

    ISSTABLE = all(real(eig(A))<0)

    LIN = @(x,u,p) A*x + B*u + p;
    OUT = @(x,u,p) C*x;

    ADJ = @(x,u,p) A'*x + C'*u + p;
    AUT = @(x,u,p) B'*x;

    Y = rk2(LIN,OUT,T,X,U,rand(P,1));

    QZ = zeros(P,1);
    Q0 = rand(P,100);
    Q1 = ones(N,1)*[0.25,0.5,0.75,1.0];
    Q2 = [zeros(N,1),ones(N,1)];

%% GRAMIAN EQUALITY

    TS = [0.01,0.005,0.001,0.0005,0.0001];

    for I=1:numel(TS)
        T1 = [TS(I),T(2)];

        WC = emgr(LIN,OUT,[J,N,O],T1,'c',QZ);
        Wo = emgr(ADJ,AUT,[J,N,O],T1,'c',QZ);
        WO = emgr(LIN,OUT,[J,N,O],T1,'o',QZ);
        WX = emgr(LIN,OUT,[J,N,O],T1,'x',QZ);
        WY = emgr(LIN,ADJ,[J,N,O],T1,'y',QZ);

        ERR01(1,I) = norm(WC-Wo,'fro')/norm(Wo,'fro');
        ERR01(2,I) = norm(Wo-WY,'fro')/norm(WY,'fro');
        ERR01(3,I) = norm(WC-WO,'fro')/norm(WO,'fro');
        ERR01(4,I) = norm(WO-WX,'fro')/norm(WX,'fro');
        ERR01(5,I) = norm(WC-WX,'fro')/norm(WX,'fro');

        fprintf('|');
    end
    fprintf('\n');

    save('../../data/ch5emgr.mat','ERR01','N','P');
%
    wa = sylvester(A,A',-B*B');

    TE = [1,2,4,8,10];

    for I=1:numel(TE)
        T2 = [T(1),TE(I)];

        WC = emgr(LIN,OUT,[J,N,O],T2,'c',QZ);
        WO = emgr(LIN,OUT,[J,N,O],T2,'o',QZ);
        WX = emgr(LIN,OUT,[J,N,O],T2,'x',QZ);

        ERR02(1,I) = norm(WC-wa,'fro')/norm(wa,'fro');
        ERR02(2,I) = norm(WO-wa,'fro')/norm(wa,'fro');
        ERR02(3,I) = norm(WX-wa,'fro')/norm(wa,'fro');

        fprintf('|');
    end
    fprintf('\n');

    save('../../data/ch5emgr.mat','ERR02','-append');


%% STATE-SPACE REDUCTION

    %% Balanced Truncation
    o = tic;
    WC = emgr(LIN,OUT,[J,N,O],T,'c',QZ);
    WO = emgr(LIN,OUT,[J,N,O],T,'o',QZ);
    [VX,DX,UX] = balancing(WC,WO); DX = diag(DX);
    OFFLINE_CO = toc(o)
    ERR11 = serror(@rk2,LIN,OUT,T,X,U,QZ,UX,VX,1);
    BND11 = sbound(@rk2,LIN,OUT,T,X,U,QZ,UX,DX,VX,1,B,C);
    save('../../data/ch5emgr.mat','ERR11','BND11','-append');

    %% Linear Balanced Truncation
    o = tic;
    WC = emgr(LIN,OUT,[J,N,O],T,'c',QZ);
    Wo = emgr(ADJ,AUT,[J,N,O],T,'c',QZ);
    [VX,DX,UX] = balancing(WC,Wo); DX = diag(DX);
    OFFLINE_Co = toc(o)
    ERR12 = serror(@rk2,LIN,OUT,T,X,U,QZ,UX,VX,1);
    BND12 = sbound(@rk2,LIN,OUT,T,X,U,QZ,UX,DX,VX,1,B,C);
    save('../../data/ch5emgr.mat','ERR12','BND12','-append');

    %% Linear Direct Truncation
    o = tic;
    WY = emgr(LIN,ADJ,[J,N,O],T,'y',QZ);
    [UX,DX,VX] = svd(WY); DX = diag(DX);
    OFFLINE_Y = toc(o)
    ERR13 = serror(@rk2,LIN,OUT,T,X,U,QZ,UX,UX',1);
    BND13 = sbound(@rk2,LIN,OUT,T,X,U,QZ,UX,DX,UX',1,B,C);
    save('../../data/ch5emgr.mat','ERR13','BND13','-append');

    %% Direct Truncation
    o = tic;
    WX = emgr(LIN,OUT,[J,N,O],T,'x',QZ);
    [UX,DX,VX] = svd(WX); DX = diag(DX);
    OFFLINE_X = toc(o)
    ERR14 = serror(@rk2,LIN,OUT,T,X,U,QZ,UX,UX',1);
    BND14 = sbound(@rk2,LIN,OUT,T,X,U,QZ,UX,DX,UX',1,B,C);
    save('../../data/ch5emgr.mat','ERR14','BND14','-append');

    %% Balanced Gains Direct Truncation
    o = tic;
    d = abs(sum((UX'*B).*(C*UX)',2)).*DX;
    [TMP,IX] = sort(d,'descend');
    for r = 1:size(A,1), UG(:,r) = UX(:,IX(r)); DG(r) = DX(IX(r)); end;
    OFFLINE_x = OFFLINE_X + toc(o)
    ERR15 = serror(@rk2,LIN,OUT,T,X,U,QZ,UG,UG',1);
    BND15 = sbound(@rk2,LIN,OUT,T,X,U,QZ,UG,DG,UG',1,B,C);
    save('../../data/ch5emgr.mat','ERR15','BND15','-append');

    %% Non-Symmetric Direct Truncation
    o = tic;
    WZ = emgr(LIN,OUT,[J,N,O],T,'x',QZ,[0,0,0,0,0,0,1,0,0,0,0,0]);
    [UX,DX,VX] = svd(WZ); DX = diag(DX);
    OFFLINE_Z = toc(o)
    ERR16 = serror(@rk2,LIN,OUT,T,X,U,QZ,UX,UX',1);
    BND16 = sbound(@rk2,LIN,OUT,T,X,U,QZ,UX,DX,UX',1,B,C);
    save('../../data/ch5emgr.mat','ERR16','BND16','-append');

%% PARAMETRIC STATE-SPACE REDUCTION

    %% Balanced Truncation
    o = tic;
    WC = emgr(LIN,OUT,[J,N,O],T,'c',Q1);
    WO = emgr(LIN,OUT,[J,N,O],T,'o',Q1);
    [VX,DX,UX] = balancing(WC,WO);
    OFFLINE_pCO = toc(o)
    ERR21 = serror(@rk2,LIN,OUT,T,X,U,Q0,UX,VX,1);
    save('../../data/ch5emgr.mat','ERR21','-append');

    %% Linear Balanced Truncation
    o = tic;
    WC = emgr(LIN,OUT,[J,N,O],T,'c',Q1);
    Wo = emgr(ADJ,AUT,[J,N,O],T,'c',Q1);
    [VX,DX,UX] = balancing(WC,Wo);
    OFFLINE_pCo = toc(o)
    ERR22 = serror(@rk2,LIN,OUT,T,X,U,Q0,UX,VX,1);
    save('../../data/ch5emgr.mat','ERR22','-append');

    %% Linear Direct Truncation
    o = tic;
    WY = emgr(LIN,ADJ,[J,N,O],T,'y',Q1);
    [UX,DX,VX] = svd(WY);
    OFFLINE_pY = toc(o)
    ERR23 = serror(@rk2,LIN,OUT,T,X,U,Q0,UX,UX',1);
    save('../../data/ch5emgr.mat','ERR23','-append');

    %% Direct Truncation
    o = tic;
    WX = emgr(LIN,OUT,[J,N,O],T,'x',Q1);
    [UX,DX,VX] = svd(WX);
    OFFLINE_pX = toc(o)
    ERR24 = serror(@rk2,LIN,OUT,T,X,U,Q0,UX,UX',1);
    save('../../data/ch5emgr.mat','ERR24','-append');

    %% Balanced Gains Direct Truncation
    o = tic;
    d = abs(sum((UX'*B).*(C*UX)',2)).*DX;
    [TMP,IX] = sort(d,'descend');
    for r = 1:size(A,1), UG(:,r) = UX(:,IX(r)); DG(r) = DX(IX(r)); end;
    OFFLINE_px = OFFLINE_pX + toc(o)
    ERR25 = serror(@rk2,LIN,OUT,T,X,U,Q0,UG,UG',1);
    save('../../data/ch5emgr.mat','ERR25','-append');

    %% Non-Symmetric Direct Truncation
    o = tic;
    WZ = emgr(LIN,OUT,[J,N,O],T,'x',Q1,[0,0,0,0,0,0,1,0,0,0,0,0]);
    [UX,DX,VX] = svd(WZ);
    OFFLINE_pZ = toc(o)
    ERR26 = serror(@rk2,LIN,OUT,T,X,U,Q0,UX,UX',1);
    save('../../data/ch5emgr.mat','ERR26','-append');

%% PARAMETER-SPACE REDUCTION

    %% Sensitivity Gramian
    o = tic;
    WS = emgr(LIN,OUT,[J,N,O],T,'s',Q2);
    [UP,DP,VP] = svd(full(WS{2}));
    OFFLINE_S = toc(o)
    ERR31 = perror(@rk2,LIN,OUT,T,X,U,Q0,UP,1);
    save('../../data/ch5emgr.mat','ERR31','-append');

    %% Centered Sensitivity Gramian
    o = tic;
    Ws = emgr(LIN,OUT,[J,N,O],T,'s',Q2,[0,0,0,0,0,0,0,0,0,0,1,0]);
    [UP,DP,VP] = svd(full(Ws{2}));
    OFFLINE_s = toc(o)
    ERR32 = perror(@rk2,LIN,OUT,T,X,U,Q0,UP,1);
    save('../../data/ch5emgr.mat','ERR32','-append');

    %% Identifiability Gramian
    o = tic;
    WI = emgr(LIN,OUT,[J,N,O],T,'i',Q2,[0,0,0,0,0,0,0,0,0,0,1,0]);
    [UP,DP,VP] = svd(WI{2});
    OFFLINE_I = toc(o)
    ERR33 = perror(@rk2,LIN,OUT,T,X,U,Q0,UP,1);
    save('../../data/ch5emgr.mat','ERR33','-append');

    %% Approximate Identifiability Gramian
    o = tic;
    Wi = emgr(LIN,OUT,[J,N,O],T,'i',Q2);
    [UP,DP,VP] = svd(Wi{2});
    OFFLINE_i = toc(o)
    ERR34 = perror(@rk2,LIN,OUT,T,X,U,Q0,UP,1);
    save('../../data/ch5emgr.mat','ERR34','-append');

    %% Cross-Identifiability Gramian
    o = tic;
    WJ = emgr(LIN,OUT,[J,N,O],T,'j',Q2,[0,0,0,0,0,0,0,0,0,0,1,0]);
    [UP,DP,VP] = svd(WJ{2});
    OFFLINE_J = toc(o)
    ERR35 = perror(@rk2,LIN,OUT,T,X,U,Q0,UP,1);
    save('../../data/ch5emgr.mat','ERR35','-append');

    %% Approximate Cross-Identifiability Gramian
    o = tic;
    Wj = emgr(LIN,OUT,[J,N,O],T,'j',Q2);
    [UP,DP,VP] = svd(Wj{2});
    OFFLINE_j = toc(o)
    ERR36 = perror(@rk2,LIN,OUT,T,X,U,Q0,UP,1);
    save('../../data/ch5emgr.mat','ERR36','-append');

%% COMBINED REDUCTION

    %% Controllability-Based Combined Reduction
    o = tic;
    Ws = emgr(LIN,OUT,[J,N,O],T,'s',Q2,[0,0,0,0,0,0,0,0,0,0,1,0]);
    WO = emgr(LIN,OUT,[J,N,O],T,'o',Q1);
    [VX,DX,UX] = balancing(WS{1},WO);
    [UP,DP,VP] = svd(full(Ws{2}));
    OFFLINE_CS = toc(o)
    ERR41 = cerror(@rk2,LIN,OUT,T,X,U,Q0,UX,VX,UP,10);
    ERR41(ERR41>1.0) = 1.0;
    save('../../data/ch5emgr.mat','ERR41','-append');

    %% Observability-Based Combined Reduction
    o = tic;
    WC = emgr(LIN,OUT,[J,N,O],T,'c',Q1);
    WI = emgr(LIN,OUT,[J,N,O],T,'i',Q2);
    [VX,DX,UX] = balancing(WC,WI{1});
    [UP,DP,VP] = svd(WI{2});
    OFFLINE_CI = toc(o)
    ERR42 = cerror(@rk2,LIN,OUT,T,X,U,Q0,UX,VX,UP,10);
    ERR42(ERR42>1.0) = 1.0;
    save('../../data/ch5emgr.mat','ERR42','-append');

    %% Cross-Gramian-Based Combined Reduction
    o = tic;
    Wj = emgr(LIN,OUT,[J,N,O],T,'j',Q2,[0,0,0,0,0,0,0,0,0,0,1,0]);
    [UX,DX,VX] = svd(Wj{1});
    [UP,DP,VP] = svd(Wj{2});
    OFFLINE_CJ = toc(o)
    ERR43 = cerror(@rk2,LIN,OUT,T,X,U,Q0,UX,UX',UP,10);
    ERR43(ERR43>1.0) = 1.0;
    save('../../data/ch5emgr.mat','ERR43','-append');
end
