function bayinv()
% mri (Chapter 7 Bayesian Inversion)
% by Christian Himpe, 2015-2016 ( wwwmath.uni-muenster.de/u/himpe )
% released under BSD 2-Clause License ( opensource.org/licenses/BSD-2-Clause )
%*
    addpath('../');

    rand('seed',1009);
    randn('seed',1009);

%% COMMON SETUP

    n = 16;
    J = 1;
    O = n;
    P = n*n;

    a = full(sprand(n,n,1.0/n));
    b = full(sprand(n,J,1.0/n));

    v = ones(P,1); v(1:n+1:end) = 0.01;

    Q0 = ones(P,1);
    QD = exp(0.25*bsxfun(@times,randn(n*n,1),v));

%% SETUP MRI

    N = 5*n;
    X = [zeros(2*n,1);ones(3*n,1)];

    T = [0.1,20.0];
    L = floor(T(2)/T(1)) + 1;
    U = [ones(J,1)./T(1),zeros(J,L)];

    A = a;
    A(1:n+1:end) = -2.0;

    [NON,NOU,LIN,OUT] = mri_util(n,A,b);

    YY = rk2(NON,NOU,T,X,U,QD);
    YD = YY + whitenoise(YY,10);

    YDmYY = norm(YD-YY,2)
    QDmQ0 = norm(QD-Q0,2)/norm(QD,2)

    save('../../data/ch7bayinv.mat','YD','QD');

    b2 = 1e-4;

%% MAIN MRI

    % FOM
    o = tic;
    Q1 = fminunc(@(q) norm(YD-rk2(NON,NOU,T,X,U,exp(q)),2)^2 + b2*norm(Q0-exp(q),2)^2,zeros(n*n,1)); Q1', Q1 = exp(Q1);
    TIM11 = toc(o)
    ERP01 = norm((QD~=1).*(Q1 - QD),2)
    ERY01 = norm(YD-rk2(NON,NOU,T,X,U,Q1),2)
    SAM01 = sample(@rk2,NON,NOU,T,X,U,Q1);
    save('../../data/ch7bayinv.mat','TIM11','ERP01','ERY01','SAM01','Q1','-append');

    v1 = 0.5*ones(P,1); v(1:n+1:end) = 0.95;
    v2 = 1.5*ones(P,1); v(1:n+1:end) = 1.05;
    rg = 45;

    % GRAMIAN-BASED ROM
    o = tic;
    WJ = emgr(NON,NOU,[J,N,O],T,'j',[v1,v2],[2,0,0,3,0,0,1,0,1,1,0,0],1.0,0,X,1.0,0.9);
    [UX,DX,vx] = svd(WJ{1}+WJ{1}'); vx = vx(:,1:rg);
    [UI,DI,vi] = svd(WJ{2}); UI = UI(:,1:rg);
    TIM02 = toc(o)

    o = tic;
    non = @(x,u,p) vx'*NON(vx*x,u,p); nou = @(x,u,p) NOU(vx*x,u,p); x = vx'*X;
    Q2 = fminunc(@(q) norm(YD-rk2(non,nou,T,x,U,exp(UI*q)),2)^2 + b2*norm(Q0-exp(UI*q),2)^2,UI'*zeros(n*n,1)); Q2', Q2 = exp(UI*Q2);
    TIM12 = toc(o)
    ERP02 = norm((QD~=1).*(Q2 - QD),2)
    ERY02 = norm(YD-rk2(NON,NOU,T,X,U,Q2),2)
    SAM02 = sample(@rk2,NON,NOU,T,X,U,Q2);
    save('../../data/ch7bayinv.mat','TIM02','TIM12','ERP02','ERY02','SAM02','Q2','-append');

    % OPTIMIZATION-BASED ROM
    ro = 25;

    o = tic;
    XP = optmor(NON,NOU,[J,N,O],T,ro,Q0,[0,0,1e-1,0,0,0],1,X,1.0,YD);
    UO = XP{1};
    UP = XP{2};
    TIM03 = toc(o)

    o = tic;
    non = @(x,u,p) UO'*NON(UO*x,u,p); nou = @(x,u,p) NOU(UO*x,u,p); x = UO'*X;
    Q3 = fminunc(@(q) norm(YD-rk2(non,nou,T,x,U,exp(UP*q)),2)^2 + b2*norm(Q0-exp(UP*q),2)^2,UP'*zeros(n*n,1)); Q3', Q3 = exp(UP*Q3);
    TIM13 = toc(o)
    ERP03 = norm((QD~=1).*(Q3 - QD),2)
    ERY03 = norm(YD-rk2(NON,NOU,T,X,U,Q3),2)
    SAM03 = sample(@rk2,NON,NOU,T,X,U,Q3);
    save('../../data/ch7bayinv.mat','TIM03','TIM13','ERP03','ERY03','SAM03','Q3','-append');

    % OPTIMIZATION-BASED ROM (DD)
    ro = 25;

    o = tic;
    XP = optmor(NON,NOU,[J,N,O],T,ro,Q0,[0,0,1e-1,1e-8,0,0],1,X,1.0,YD);
    UO = XP{1};
    UP = XP{2};
    TIM04 = toc(o)

    o = tic;
    non = @(x,u,p) UO'*NON(UO*x,u,p); nou = @(x,u,p) NOU(UO*x,u,p); x = UO'*X;
    Q4 = fminunc(@(q) norm(YD-rk2(non,nou,T,x,U,exp(UP*q)),2)^2 + b2*norm(Q0-exp(UP*q),2)^2,UP'*zeros(n*n,1)); Q4', Q4 = exp(UP*Q4);
    TIM14 = toc(o)
    ERP04 = norm((QD~=1).*(Q4 - QD),2)
    ERY04 = norm(YD-rk2(NON,NOU,T,X,U,Q4),2)
    SAM04 = sample(@rk2,NON,NOU,T,X,U,Q4);
    save('../../data/ch7bayinv.mat','TIM04','TIM14','ERP04','ERY04','SAM04','Q4','-append');
end
